from tests.utils import *
import networkx as nx

def test_label_simple():
    cfg = parse_and_create_cfg("""int main()
{
    int x = 0;
end:
    x = 10;
    return x;
}
""")
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 5)
    assert nx.is_directed_acyclic_graph(cfg)

    FUNC_ENTRY_node = get_node_by_label(cfg, "FUNC_ENTRY")
    x_10_node = get_node_by_code(cfg, "x = 10;")
    assert len(list(nx.all_simple_paths(cfg, FUNC_ENTRY_node, x_10_node))) > 0


def test_goto_label_simple():
    cfg = parse_and_create_cfg("""int main()
{
    int x = 0;
    goto end;
end:
    x = 10;
    return x;
}
""")
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (7, 6)
    assert nx.is_directed_acyclic_graph(cfg)

    FUNC_ENTRY_node = get_node_by_label(cfg, "FUNC_ENTRY")
    goto_node = get_node_by_code(cfg, "goto end;")
    label_node = get_node_by_code(cfg, "end:")
    x_10_node = get_node_by_code(cfg, "x = 10;")
    assert get_adj_label(cfg, goto_node, label_node) == "goto"
    assert len(list(nx.all_simple_paths(cfg, FUNC_ENTRY_node, x_10_node))) > 0


def test_goto_():
    cfg = parse_and_create_cfg("""int main()
{
    int x = 0;
    goto end;
    x = 20;
end:
    x = 10;
    return x;
}
""")
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (8, 7)
    assert nx.is_directed_acyclic_graph(cfg)

    FUNC_ENTRY_node = get_node_by_label(cfg, "FUNC_ENTRY")
    goto_node = get_node_by_code(cfg, "goto end;")
    x_20_node = get_node_by_code(cfg, "x = 20;")
    x_10_node = get_node_by_code(cfg, "x = 10;")
    assert not any(x_20_node in p for p in nx.all_simple_paths(cfg, goto_node, x_10_node))
    assert list(nx.all_simple_paths(cfg, FUNC_ENTRY_node, x_20_node)) == []



from tree_climber.tree_sitter_utils import c_parser


class BaseVisitor:
    """
    Extensible AST visitor.
    Traverses and prints all nodes recursively by default.
    """

    def visit(self, n, **kwargs):
        return getattr(self, f"visit_{n.type}", self.visit_default)(n=n, **kwargs)

    def visit_children(self, n, **kwargs):
        for i, c in enumerate(n.children):
            should_continue = self.visit(c, child_idx=i, **kwargs)
            if should_continue == False:
                break

    def visit_default(self, n, **kwargs):
        print("enter", n, "kwargs", kwargs)
        self.visit_children(n)
        print("exit", n)


def test_print_ast():
    with open("tests/data/do_while_continue.c", "rb") as f:
        tree = c_parser.parse(f.read())

    v = BaseVisitor()
    v.visit(tree.root_node)



"""
Global configuration
"""

import os

tree_climber_DATAROOT = os.path.expanduser("~/.tree_climber/")
TREE_SITTER_LIB_PREFIX = os.environ.get("tree_climber__TREE_SITTER_LIB_PREFIX", os.path.join(tree_climber_DATAROOT, "lib"))
DRAW_CFG = os.environ.get("tree_climber__DRAW_CFG", False)



"""
Output CPG representation.
AST annotated with CFG, DUC edges.
"""

import networkx as nx


def make_cpg(ast, cfg, duc):
    ast = ast.copy()
    cfg = cfg.copy()
    duc = duc.copy()
    nx.set_edge_attributes(ast, {(u, v): "AST" for u, v in ast.edges()}, "graph_type")
    nx.set_edge_attributes(
        cfg, {(u, v, k): "CFG" for u, v, k in cfg.edges(keys=True)}, "graph_type"
    )
    nx.set_edge_attributes(duc, {(u, v): "DUC" for u, v in duc.edges()}, "graph_type")
    max_ast_node = max(ast.nodes())
    cfg = nx.relabel_nodes(
        cfg,
        {
            n: attr.get("ast_node", max_ast_node + n + 1)
            for n, attr in cfg.nodes(data=True)
        },
    )
    duc = nx.relabel_nodes(
        duc, {n: attr.get("ast_node") for n, attr in duc.nodes(data=True)}
    )
    cpg = nx.MultiDiGraph(ast)
    cpg = nx.compose(cpg, cfg)
    cpg = nx.compose(cpg, nx.MultiDiGraph(duc))

    return cpg



import os
from tree_climber.ast_creator import ASTCreator
from tree_climber.base_visitor import BaseVisitor
from tree_climber.cfg_creator import CFGCreator
from tree_climber.config import DRAW_CFG
from tree_climber.tree_sitter_utils import c_parser
import networkx as nx
import matplotlib.pyplot as plt

def draw(cfg, dataflow_solution=None, ax=None):
    pos = nx.nx_pydot.graphviz_layout(cfg, prog="dot")
    nx.draw_networkx_nodes(cfg, pos=pos, ax=ax)
    double_edges = [(u, v, attr.get("label", "")) for u, v, attr in cfg.edges(data=True) if u in cfg.successors(v) and v in cfg.successors(u)]
    single_edges = [(u, v, attr.get("label", "")) for u, v, attr in cfg.edges(data=True) if (u, v, attr.get("label", "")) not in double_edges]
    nx.draw_networkx_edges(cfg, pos=pos, edgelist=[(u, v) for u, v, label in single_edges], ax=ax)
    nx.draw_networkx_edge_labels(cfg, pos=pos, edge_labels={(u, v): label for u, v, label in single_edges})
    nx.draw_networkx_edges(cfg, pos=pos, edgelist=[(u, v) for u, v, label in double_edges], connectionstyle=f"arc3,rad=0.1")
    nx.draw_networkx_edge_labels(cfg, pos=pos, edge_labels={(u, v): label for u, v, label in double_edges})#, connectionstyle=f"arc3,rad=0.1")
    nx.draw_networkx_labels(cfg, pos=pos, labels={n: attr.get("label", "<NO LABEL>") for n, attr in cfg.nodes(data=True)}, ax=ax)
    if dataflow_solution is not None:
        nx.draw_networkx_labels(cfg, pos=pos, font_color="r", labels={n: "\n\n\n{" + ", ".join(dataflow_solution[n]) + "}" if dataflow_solution[n] else "" for n in cfg.nodes()}, ax=ax)
    plt.show()


def parse_and_create_cfg(code, print_ast=False, draw_cfg=bool(DRAW_CFG)):
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    if print_ast:
        pos = nx.drawing.nx_agraph.graphviz_layout(ast, prog='dot')
        nx.draw(ast, pos=pos, labels={n: attr["label"] for n, attr in ast.nodes(data=True)}, with_labels = True)
        plt.show()
    cfg = CFGCreator.make_cfg(ast)
    if draw_cfg:
        draw(cfg)
    return cfg

def get_adj_label(cfg, u, v):
    """get label of first edge connecting u and v in cfg"""
    return list(cfg.adj[u][v].values())[0].get("label", "<NO LABEL>")

def get_node_by_code(cfg, code, get="first"):
    matches = (n for n, attr in cfg.nodes(data=True) if code == attr.get("code", "<NO CODE>"))
    if get == "first":
        return next(matches)
    elif get == "all":
        return list(matches)
    else:
        raise NotImplementedError(get)

def get_node_by_label(cfg, label):
    return next(n for n, attr in cfg.nodes(data=True) if label == attr.get("label", "<NO LABEL>"))



from tests.utils import *
from tree_climber.dataflow.reaching_def import ReachingDefinitionSolver
import pytest


def test_solve():
    code = """int main()
    {
        int x = 0;
        if (true) {
            x += 5;
        }
        printf("%d\\n", x);
        x = 10;
        return x;
    }
    """
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    cfg = CFGCreator.make_cfg(ast)
    solver = ReachingDefinitionSolver(cfg)
    solution_in, solution_out = solver.solve()
    assert solution_out[get_node_by_code(cfg, "return x;")] == {2}
    assert solution_out[get_node_by_code(cfg, """printf("%d\\n", x);""")] == {0, 1}


@pytest.mark.slow
def test_debug():
    code = """int main()
    {
        int x = 0;
        if (true) {
            x += 5;
        }
        printf("%d\\n", x);
        x = 10;
        return x;
    }
    """
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    cfg = CFGCreator.make_cfg(ast)
    solver = ReachingDefinitionSolver(cfg, verbose=1)
    solution_in, solution_out = solver.solve()
    draw(
        cfg,
        {
            n: sorted(set(map(solver.def2code.__getitem__, b)))
            for n, b in solution_out.items()
        },
    )
    print(solution_out)



from pathlib import Path

this_file = Path(__file__)
source_dir = this_file.parent
root_dir = source_dir.parent
data_dir = root_dir / "tests/data"

example_c = data_dir / "example.c"
assert example_c.exists(), example_c



from tests.utils import *

from matplotlib import pyplot as plt
from tree_climber.globals import example_c
from tree_climber.ast_creator import ASTCreator
from tree_climber.dataflow.def_use import make_duc
from tree_climber.export.cpg import make_cpg
from tree_climber.tree_sitter_utils import c_parser
from tree_climber.cfg_creator import CFGCreator
import pytest


@pytest.mark.slow
def test_debug():
    code = example_c.read_text()
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    cfg = CFGCreator.make_cfg(ast)
    duc = make_duc(cfg)
    cpg = make_cpg(ast, cfg, duc)

    pos = nx.nx_pydot.graphviz_layout(cpg, prog="dot")
    nx.draw(cpg, pos=pos)
    nx.draw_networkx_labels(
        cpg,
        pos=pos,
        labels={n: attr.get("label", "<NO LABEL>") for n, attr in cpg.nodes(data=True)},
    )
    for graph_type, color in {
        "AST": "black",
        "CFG": "blue",
        "DUC": "red",
    }.items():
        nx.draw_networkx_edges(
            cpg,
            pos=pos,
            edge_color=color,
            edgelist=[
                (u, v)
                for u, v, k, attr in cpg.edges(keys=True, data=True)
                if attr["graph_type"] == graph_type
            ],
        )
    nx.draw_networkx_edge_labels(
        cpg,
        pos=pos,
        edge_labels={
            (u, v): attr.get("label", "")
            for u, v, k, attr in cpg.edges(keys=True, data=True)
        },
    )
    plt.show()



from collections import defaultdict
from matplotlib import pyplot as plt
from tree_climber.ast_creator import ASTCreator
from tree_climber.base_visitor import BaseVisitor
import networkx as nx
from tree_climber.tree_sitter_utils import c_parser


class CFGCreator(BaseVisitor):
    """
    AST visitor which creates a CFG.
    After traversing the AST by calling visit() on the root, self.cfg has a complete CFG.
    """

    def __init__(self, ast):
        super(CFGCreator).__init__()
        self.ast = ast
        self.cfg = nx.MultiDiGraph()
        self.node_id = 0
        self.fringe = []
        self.break_fringe = []
        self.continue_fringe = []
        self.gotos = {}
        self.labels = {}

    @staticmethod
    def make_cfg(ast):
        visitor = CFGCreator(ast)
        visitor.visit(ast.graph["root_node"])
        cfg = visitor.cfg
        # Postprocessing

        # pass through dummy nodes
        nodes_to_remove = []
        for n, attr in cfg.nodes(data=True):
            if attr.get("dummy", False):
                preds = list(cfg.predecessors(n))
                succs = list(cfg.successors(n))
                # Forward label from edges incoming to dummy.
                for pred in preds:
                    new_edge_label = list(cfg.adj[pred][n].values())[0].get(
                        "label", None
                    )
                    for succ in succs:
                        cfg.add_edge(pred, succ, label=new_edge_label)
                nodes_to_remove.append(n)
        cfg.remove_nodes_from(nodes_to_remove)

        return visitor.cfg

    def get_children(self, n):
        return list(self.ast.successors(n))

    def visit(self, n, **kwargs):
        return getattr(
            self, "visit_" + self.ast.nodes[n]["node_type"], self.visit_default
        )(n=n, **kwargs)

    def visit_children(self, n, **kwargs):
        for c in self.ast.successors(n):
            should_continue = self.visit(c, **kwargs)
            if should_continue == False:
                break

    def add_dummy_node(self):
        """dummy nodes are nodes whose connections should be forwarded in a post-processing step"""
        node_id = self.node_id
        self.cfg.add_node(node_id, dummy=True, label="DUMMY")
        self.node_id += 1
        return node_id

    def add_cfg_node(self, ast_node, label=None, **kwargs):
        node_id = self.node_id
        ast_kwargs = {}
        if ast_node is not None:
            attr = self.ast.nodes[ast_node]
            ast_kwargs.update(attr)
        if label is not None:
            ast_kwargs["label"] = label
        if ast_node is not None:
            ast_kwargs["ast_node"] = ast_node
        ast_kwargs.update(kwargs)
        self.cfg.add_node(node_id, **ast_kwargs)
        self.node_id += 1
        return node_id

    def add_edge_from_fringe_to(self, node_id):
        # Assign edges with labels
        fringe_by_type = defaultdict(list)
        for source in self.fringe:
            if isinstance(source, tuple):
                fringe_by_type[source[1]].append(source[0])
            else:
                fringe_by_type[None].append(source)
        for edge_type, fringe in fringe_by_type.items():
            kwargs = {}
            if edge_type is not None:
                kwargs["label"] = str(edge_type)
            self.cfg.add_edges_from(zip(fringe, [node_id] * len(self.fringe)), **kwargs)
        self.fringe = []

    """
    VISITOR RULES
    """

    def visit_function_definition(self, n, **kwargs):
        entry_id = self.add_cfg_node(None, "FUNC_ENTRY")
        self.cfg.graph["entry"] = entry_id
        self.add_edge_from_fringe_to(entry_id)
        self.fringe.append(entry_id)
        self.visit_children(n, **kwargs)
        exit_id = self.add_cfg_node(None, "FUNC_EXIT")
        self.add_edge_from_fringe_to(exit_id)
        # paste goto edges
        for label in self.gotos:
            try:
                self.cfg.add_edge(self.gotos[label], self.labels[label], label="goto")
            except KeyError:
                print("WARNING: missing goto target. Skipping.", f"label={label}", f"gotos={self.gotos}")
        for n in nx.descendants(self.cfg, entry_id):
            attr = self.cfg.nodes[n]
            if attr.get("n", None) is not None and attr["n"].type == "return_statement":
                self.cfg.add_edge(n, exit_id, label="return")
        self.fringe.append(exit_id)

    def visit_default(self, n, **kwargs):
        self.visit_children(n)

    """STRAIGHT LINE STATEMENTS"""

    def enter_statement(self, n):
        node_id = self.add_cfg_node(n)
        self.add_edge_from_fringe_to(node_id)
        self.fringe.append(node_id)

    def visit_expression_statement(self, n, **kwargs):
        self.enter_statement(n)
        self.visit_default(n, **kwargs)

    def visit_declaration(self, n, **kwargs):
        self.enter_statement(n)
        self.visit_default(n, **kwargs)

    """STRUCTURED CONTROL FLOW"""

    def visit_if_statement(self, n, **kwargs):
        condition = self.get_children(self.get_children(n)[0])[0]
        condition_id = self.add_cfg_node(condition)
        self.add_edge_from_fringe_to(condition_id)
        self.fringe.append((condition_id, True))

        compound_statement = self.get_children(n)[1]
        self.visit(compound_statement)
        # NOTE: this assert doesn't work in the case of an if with an empty else
        # assert len(self.fringe) == 1, "fringe should now have last statement of compound_statement"

        if len(self.get_children(n)) > 2:
            else_compound_statement = self.get_children(n)[2]
            old_fringe = self.fringe
            self.fringe = [(condition_id, False)]
            self.visit(else_compound_statement)
            self.fringe = old_fringe + self.fringe
        else:
            self.fringe.append((condition_id, False))

    def visit_for_statement(self, n, **kwargs):
        n_attr = self.ast.nodes[n]
        # print(n, n_attr["node_type"], kwargs, n_attr)
        i = 0
        if n_attr.get("has_init", False):
            init = self.get_children(n)[i]
            i += 1
        else:
            init = None
        if n_attr.get("has_cond", False):
            cond = self.get_children(n)[i]
            i += 1
        else:
            cond = None
        if n_attr.get("has_incr", False):
            incr = self.get_children(n)[i]
            i += 1
        else:
            incr = None

        if cond is not None:
            cond_id = self.add_cfg_node(cond)
        else:
            cond_id = self.add_cfg_node(None, f"<TRUE>")

        if init is not None:
            init_id = self.add_cfg_node(init)
            self.add_edge_from_fringe_to(init_id)
            self.cfg.add_edge(init_id, cond_id)
        else:
            self.add_edge_from_fringe_to(cond_id)
        self.fringe.append((cond_id, True))

        compound_statement = self.get_children(n)[i]
        self.visit(compound_statement)
        # NOTE: this assert doesn't work in the case of an if with an empty else
        # assert len(self.fringe) == 1, "fringe should now have last statement of compound_statement"
        if incr is not None:
            incr_id = self.add_cfg_node(incr)
            self.add_edge_from_fringe_to(incr_id)
            self.cfg.add_edge(incr_id, cond_id)
            self.cfg.add_edges_from(
                zip(self.continue_fringe, [incr_id] * len(self.continue_fringe)),
                label="continue",
            )
            self.continue_fringe = []
        else:
            self.add_edge_from_fringe_to(cond_id)
            self.cfg.add_edges_from(
                zip(self.continue_fringe, [cond_id] * len(self.continue_fringe)),
                label="continue",
            )
            self.continue_fringe = []
        self.fringe.append((cond_id, False))

        self.fringe += [(n, "break") for n in self.break_fringe]
        self.break_fringe = []

    def visit_while_statement(self, n, **kwargs):
        cond = self.get_children(self.get_children(n)[0])[0]
        cond_id = self.add_cfg_node(cond)

        self.add_edge_from_fringe_to(cond_id)
        self.fringe.append((cond_id, True))

        compound_statement = self.get_children(n)[1]
        self.visit(compound_statement)
        self.add_edge_from_fringe_to(cond_id)
        self.fringe.append((cond_id, False))

        self.cfg.add_edges_from(
            zip(self.continue_fringe, [cond_id] * len(self.continue_fringe)),
            label="continue",
        )
        self.continue_fringe = []
        self.fringe += [(n, "break") for n in self.break_fringe]
        self.break_fringe = []

    def visit_do_statement(self, n, **kwargs):
        dummy_id = self.add_dummy_node()
        self.add_edge_from_fringe_to(dummy_id)
        self.fringe.append(dummy_id)

        compound_statement = self.get_children(n)[0]
        self.visit(compound_statement)

        cond = self.get_children(self.get_children(n)[1])[0]
        cond_id = self.add_cfg_node(cond)
        self.add_edge_from_fringe_to(cond_id)
        self.cfg.add_edge(cond_id, dummy_id, label=str(True))
        self.fringe.append((cond_id, False))

        self.cfg.add_edges_from(
            zip(self.continue_fringe, [cond_id] * len(self.continue_fringe)),
            label="continue",
        )
        self.continue_fringe = []
        self.fringe += [(n, "break") for n in self.break_fringe]
        self.break_fringe = []

    def visit_switch_statement(self, n, **kwargs):
        cond = self.get_children(self.get_children(n)[0])[0]
        cond_id = self.add_cfg_node(cond)
        self.add_edge_from_fringe_to(cond_id)
        cases = self.get_children(self.get_children(n)[1])
        default_was_hit = False
        for case in cases:
            while self.ast.nodes[case]["node_type"] != "case_statement":
                if self.ast.nodes[case]["node_type"] == "labeled_statement":
                    self.add_label_node(case)
                    case = self.get_children(case)[1]
                else:
                    raise NotImplementedError(self.ast.nodes[case]["node_type"])
            case_children = self.get_children(case)
            case_attr = self.ast.nodes[case]
            if len(self.get_children(case)) == 0:
                continue
            body_nodes = [
                c
                for c in case_children
                if case_attr["body_begin"] <= self.ast.nodes[c]["child_idx"]
            ]
            if case_attr["is_default"]:
                default_was_hit = True
            case_text = self.ast.nodes[case]["code"]
            case_text = case_text[: case_text.find(":") + 1]
            # TODO: append previous cases with no body
            self.fringe.append((cond_id, case_text))
            for body_node in body_nodes:
                should_continue = self.visit(body_node)
                if should_continue == False:
                    break
        if not default_was_hit:
            self.fringe.append((cond_id, "default:"))
        self.fringe += [(n, "break") for n in self.break_fringe]
        self.break_fringe = []

    def visit_return_statement(self, n, **kwargs):
        node_id = self.add_cfg_node(n)
        self.add_edge_from_fringe_to(node_id)
        self.visit_default(n, **kwargs)
        # This is meant to skip adding subsequent statements to the CFG.
        # TODO: consider how to handle this with goto statements.
        return False

    def visit_break_statement(self, n, **kwargs):
        node_id = self.add_cfg_node(n)
        self.add_edge_from_fringe_to(node_id)
        self.break_fringe.append(node_id)
        self.visit_default(n, **kwargs)
        return False

    def visit_continue_statement(self, n, **kwargs):
        node_id = self.add_cfg_node(n)
        self.add_edge_from_fringe_to(node_id)
        self.continue_fringe.append(node_id)
        self.visit_default(n, **kwargs)
        return False

    def visit_goto_statement(self, n, **kwargs):
        node_id = self.add_cfg_node(n)
        self.add_edge_from_fringe_to(node_id)
        statement_identifier_attr = self.ast.nodes[self.get_children(n)[0]]
        assert statement_identifier_attr["node_type"] == "statement_identifier"
        self.gotos[statement_identifier_attr["code"]] = node_id
        self.visit_default(n, **kwargs)

    def add_label_node(self, n):
        code = self.ast.nodes[n]["code"]
        code = code[:code.find(":")+1]
        node_id = self.add_cfg_node(n, code=code)
        self.add_edge_from_fringe_to(node_id)
        statement_identifier_attr = self.ast.nodes[self.get_children(n)[0]]
        assert statement_identifier_attr["node_type"] == "statement_identifier"
        self.labels[statement_identifier_attr["code"]] = node_id
        self.fringe.append(node_id)

    def visit_labeled_statement(self, n, **kwargs):
        self.add_label_node(n)
        self.visit_default(n, **kwargs)


def test():
    code = """int a = 30;

struct foo {
    int z;
};

int main()
{
    int x = 0;
    for (int i = 0; i < 10; i ++) {
        x -= i;
    }
    x += 20;
    switch (x) {
        case 0:
            x += 30;
            break;
        case 1:
            x -= 1;
        case 2:
            x -= a;
            break;
        default:
            return -2;
    }
    while (x > 10) {
        x -= 5;
    }
    do {
        x -= 5;
    } while (x > 0);
    return x;
}

struct foo;
"""
    tree = c_parser.parse(bytes(code, "utf8"))

    fig, ax = plt.subplots(2)
    ast = ASTCreator.make_ast(tree.root_node)
    pos = nx.drawing.nx_agraph.graphviz_layout(ast, prog="dot")
    nx.draw(
        ast,
        pos=pos,
        labels={n: attr["label"] for n, attr in ast.nodes(data=True)},
        with_labels=True,
        ax=ax[0],
    )

    cfg = CFGCreator.make_cfg(ast)
    pos = nx.drawing.nx_agraph.graphviz_layout(cfg, prog="dot")
    nx.draw(
        cfg,
        pos=pos,
        labels={n: attr["label"] for n, attr in cfg.nodes(data=True)},
        with_labels=True,
        ax=ax[1],
    )
    nx.draw_networkx_edge_labels(
        cfg,
        pos=pos,
        edge_labels={
            (u, v): attr.get("label", "") for (u, v, attr) in cfg.edges(data=True)
        },
        ax=ax[1],
    )
    plt.show()



from tests.utils import *
import networkx as nx

def test_switch_simple():
    cfg = parse_and_create_cfg("""int main()
    {
        switch(true) {
            case 0:
                return 0;
            default:
                return 1;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert nx.is_directed_acyclic_graph(cfg)

def test_switch_complex():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        switch(x) {
            case 0:
                return 0;
            case 1:
                return 1;
            case 2:
                x ++;
            default:
                x = -1;
                break;
        }
        return x;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (10, 12)
    assert nx.is_directed_acyclic_graph(cfg)

    cond_node = get_node_by_code(cfg, "x")
    return_0_node = get_node_by_code(cfg, "return 0;")
    return_1_node = get_node_by_code(cfg, "return 1;")
    x_plus_node = get_node_by_code(cfg, "x ++;")
    x_minus_node = get_node_by_code(cfg, "x = -1;")
    break_node = get_node_by_code(cfg, "break;")
    return_x_node = get_node_by_code(cfg, "return x;")

    assert get_adj_label(cfg, cond_node, return_0_node) == "case 0:"
    assert get_adj_label(cfg, cond_node, return_1_node) == "case 1:"
    assert get_adj_label(cfg, cond_node, x_plus_node) == "case 2:"
    assert get_adj_label(cfg, cond_node, x_minus_node) == "default:"
    assert get_adj_label(cfg, break_node, return_x_node) == "break"

def test_switch_nodefault():
    cfg = parse_and_create_cfg("""int main()
    {
        switch(-1) {
            case 0:
                return 0;
            case 1:
                return 1;
        }
        return 5;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 7)
    assert nx.is_directed_acyclic_graph(cfg)

    cond_node = get_node_by_code(cfg, "-1")
    return_0_node = get_node_by_code(cfg, "return 0;")
    return_1_node = get_node_by_code(cfg, "return 1;")
    return_5_node = get_node_by_code(cfg, "return 5;")

    assert get_adj_label(cfg, cond_node, return_0_node) == "case 0:"
    assert get_adj_label(cfg, cond_node, return_1_node) == "case 1:"
    assert get_adj_label(cfg, cond_node, return_5_node) == "default:"



from tree_climber.tree_sitter_utils import c_parser
from tree_climber.ast_creator import ASTCreator
from tree_climber.cfg_creator import CFGCreator
from tree_climber.dataflow.dataflow_solver import DataflowSolver
import networkx as nx


def get_definition(ast_node):
    if ast_node.type == "identifier":
        return ast_node.text.decode()
    elif ast_node.type == "pointer_declarator":
        return get_definition(ast_node.children[1])
    elif ast_node.type == "init_declarator":
        return get_definition(ast_node.children[0])
    elif ast_node.type == "declaration":
        return get_definition(ast_node.children[1])
    elif ast_node.type == "assignment_expression":
        return get_definition(ast_node.children[0])
    elif ast_node.type == "update_expression":
        return get_definition(ast_node.children[0])
    elif ast_node.type == "expression_statement":
        return get_definition(ast_node.children[0])
    return None


class ReachingDefinitionSolver(DataflowSolver):
    """
    reaching definition
    https://en.wikipedia.org/wiki/Reaching_definition
    """

    def __init__(self, cfg, verbose=0):
        super().__init__(cfg, verbose)

        node2def = {}
        def2node = {}
        id2def = {}
        def2id = {}
        def2code = {}
        def_idx = 0
        for n in cfg.nodes():
            attr = cfg.nodes[n]
            if "n" in attr:
                ast_node = attr["n"]
                _id = get_definition(ast_node)
                if _id is not None:
                    node2def[n] = def_idx
                    def2node[def_idx] = n

                    if _id not in id2def:
                        id2def[_id] = set()
                    id2def[_id].add(def_idx)
                    def2id[def_idx] = _id
                    def2code[def_idx] = ast_node.text.decode()

                    def_idx += 1
        if verbose >= 1:
            print("node2def", node2def)
            print("def2node", def2node)
            print("id2def", id2def)
            print("def2id", def2id)
            print("def2code", def2code)
        self.node2def = node2def
        self.def2node = def2node
        self.id2def = id2def
        self.def2id = def2id
        self.def2code = def2code

    def gen(self, n) -> set:
        if n in self.node2def:
            d = self.node2def[n]
            if self.verbose >= 2:
                print("gen", n, d)
            return {d}
        else:
            return set()

    def kill(self, n) -> set:
        if n in self.node2def:
            d = self.node2def[n]
            if d in self.def2id:
                i = self.def2id[d]
                if i in self.id2def:
                    if self.verbose >= 2:
                        print("kill", n, self.id2def[i])
                    return self.id2def[i]
        else:
            return set()



from tests.utils import *
import networkx as nx

def test_for_simple():
    cfg = parse_and_create_cfg("""int main()
    {
        for (int i = 0; i < 10; i ++) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 6)
    assert len(list(nx.simple_cycles(cfg))) == 1

    cond_node = get_node_by_code(cfg, "i < 10")
    x_0_node = get_node_by_code(cfg, "x = 0;")
    assert get_adj_label(cfg, cond_node, x_0_node) == "True"
    FUNC_EXIT_node = get_node_by_label(cfg, "FUNC_EXIT")
    assert get_adj_label(cfg, cond_node, FUNC_EXIT_node) == "False"

def test_for_nocompound():
    cfg = parse_and_create_cfg("""int main()
    {
        for (int i = 0; i < 10; i ++)
            x = 0;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 6)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_noinit():
    cfg = parse_and_create_cfg("""int main()
    {
        for (; i < 10; i ++) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_nocond():
    cfg = parse_and_create_cfg("""int main()
    {
        for (int i = 0; ; i++) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 6)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_noincr():
    cfg = parse_and_create_cfg("""int main()
    {
        for (int i = 0; i < 10;) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_noinitincr():
    cfg = parse_and_create_cfg("""int main()
    {
        for (; i < 10;) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_noinitcond():
    cfg = parse_and_create_cfg("""int main()
    {
        for (; ; i++) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_nocondincr():
    cfg = parse_and_create_cfg("""int main()
    {
        for (int i = 0; ; ) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_nothing():
    cfg = parse_and_create_cfg("""int main()
    {
        for (; ; ) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_nested():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        for (int i = 0; i < 10; i ++) {
            for (int j = 0; j < 10; j ++) {
                x += j;
            }
            x -= i;
        }
        return x;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (12, 13)
    assert len(list(nx.simple_cycles(cfg))) == 2

def test_for_break_all_cases():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        for (int i = 0; i < 10; i ++) {
            if (x > 5) {
                x = 22;
                break;
            }
            x = 10;
            if (i < 5) {
                return -1;
            }
            x += 5;
            break;
        }
        x = 11;
        return x;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (16, 18)
    assert nx.is_directed_acyclic_graph(cfg)
    
    cond_node = get_node_by_code(cfg, "i < 10")
    x_5_node = get_node_by_code(cfg, "x > 5")
    assert get_adj_label(cfg, cond_node, x_5_node) == "True"

    x_22_node = get_node_by_code(cfg, "x = 22;")
    assert get_adj_label(cfg, x_5_node, x_22_node) == "True"

    break_1_node, break_2_node = get_node_by_code(cfg, "break;", get="all")
    return_x_node = get_node_by_code(cfg, "return x;")
    x_11_node = get_node_by_code(cfg, "x = 11;")
    assert get_adj_label(cfg, break_1_node, x_11_node) == "break"
    assert (break_1_node, return_x_node) not in cfg.edges
    assert (break_1_node, cond_node) not in cfg.edges
    assert get_adj_label(cfg, break_2_node, x_11_node) == "break"
    assert (break_2_node, return_x_node) not in cfg.edges
    assert (break_2_node, cond_node) not in cfg.edges

def test_for_break():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        for (int i = 0; i < 10; i ++) {
            if (x > 5) {
                x = 0;
                break;
            }
            x = 10;
            if (i < 5) {
                return -1;
            }
            x += 5;
        }
        x = 10;
        return x;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (15, 17)
    assert len(list(nx.simple_cycles(cfg))) == 1

def test_for_continue():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        for (int i = 0; i < 10; i ++) {
            if (x > 5) {
                x = 22;
                continue;
            }
            x = 10;
            if (i < 5) {
                continue;
            }
            x += 5;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (13, 15)
    assert len(list(nx.simple_cycles(cfg))) == 3
    
    cond_node = get_node_by_code(cfg, "i < 10")
    x_5_node = get_node_by_code(cfg, "x > 5")
    assert get_adj_label(cfg, cond_node, x_5_node) == "True"

    x_22_node = get_node_by_code(cfg, "x = 22;")
    assert get_adj_label(cfg, x_5_node, x_22_node) == "True"

    continue_1_node, continue_2_node = get_node_by_code(cfg, "continue;", get="all")
    incr_node = get_node_by_code(cfg, "i ++")
    assert get_adj_label(cfg, continue_1_node, incr_node) == "continue"
    assert get_adj_label(cfg, continue_2_node, incr_node) == "continue"



import json
from pathlib import Path
import traceback
from tree_climber.ast_creator import ASTCreator
from tree_climber.cfg_creator import CFGCreator
from tree_climber.dataflow.def_use import make_duc
from tree_climber.export.cpg import make_cpg
from tree_climber.tree_sitter_utils import c_parser
import networkx as nx
from networkx.readwrite import json_graph
import matplotlib.pyplot as plt
import argparse
from networkx.drawing.nx_agraph import write_dot


def draw_cfg(cfg, entry=None):
    if cfg.number_of_nodes() > 1000 or cfg.number_of_edges() > 1000:
        print("fuck man I'm not drawing that!")
        return
    pos = nx.spring_layout(cfg, seed=0)
    nx.draw(cfg, pos=pos)
    nx.draw_networkx_labels(
        cfg,
        pos=pos,
        labels={n: attr.get("label", "<NO LABEL>") for n, attr in cfg.nodes(data=True)},
    )
    if entry is not None:
        plt.title(cfg.nodes[entry]["n"].text)
    plt.show()


def detect_bugs(cpg):
    # detect npd bug with reaching definition
    ast = nx.edge_subgraph(
        cpg,
        [
            (u, v, k)
            for u, v, k, attr in cpg.edges(data=True, keys=True)
            if attr["graph_type"] == "AST"
        ],
    )
    duc = nx.edge_subgraph(
        cpg,
        [
            (u, v, k)
            for u, v, k, attr in cpg.edges(data=True, keys=True)
            if attr["graph_type"] == "DUC"
        ],
    )
    null_assignment = [
        n
        for n, attr in cpg.nodes(data=True)
        if attr.get("node_type", "<NO TYPE>")
        in ("expression_statement", "init_declarator")
        and any(
            ast.nodes[m].get("node_type", "<NO TYPE>") == "null"
            for m in nx.descendants(ast, n)
        )
    ]
    for ass in null_assignment:
        for usage in duc.adj[ass]:
            usage_attr = cpg.nodes[usage]
            call_expr = next(
                (
                    ch
                    for ch in ast.adj[usage]
                    if cpg.nodes[ch]["node_type"] == "call_expression"
                ),
                None,
            )
            if call_expr is None:
                continue
            id_expr = next(
                ch
                for ch in ast.adj[call_expr]
                if cpg.nodes[ch]["node_type"] == "identifier"
            )
            id_expr_attr = cpg.nodes[id_expr]
            if id_expr_attr["code"] == "printf":
                print(
                    f"""possible npd of {next(attr["label"] for u, v, k, attr in duc.edges(data=True, keys=True) if u == ass and v == usage)} at line {id_expr_attr["start"][0]+1} column {id_expr_attr["start"][1]+1}: {usage_attr["code"]}"""
                )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="filename to parse")
    parser.add_argument(
        "--draw_ast", action="store_true", help="draw AST (Abstract Syntax Tree)"
    )
    parser.add_argument(
        "--write_ast", type=str, help="write AST (Abstract Syntax Tree) to DOT"
    )
    parser.add_argument(
        "--draw_cfg", action="store_true", help="draw CFG (Control-Flow Graph)"
    )
    parser.add_argument(
        "--write_cfg", type=str, help="write CFG (Control-Flow Graph) to DOT"
    )
    parser.add_argument(
        "--write_cfg_json", action="store_true", help="write CFG to file"
    )
    parser.add_argument(
        "--draw_duc", action="store_true", help="draw DUC (Def-Use Chain)"
    )
    parser.add_argument(
        "--draw_cpg", action="store_true", help="draw CPG (Code Property Graph)"
    )
    parser.add_argument(
        "--each_function",
        action="store_true",
        help="draw each function's CFG as a separate plot",
    )
    args = parser.parse_args()

    args.filename = Path(args.filename)
    if args.filename.is_dir():
        filenames = list(args.filename.rglob("*.c"))
    elif args.filename.is_file():
        filenames = [args.filename]
    else:
        raise FileNotFoundError(args.filename)
    print("parsing", len(filenames), "files", filenames[:5])
    for filename in filenames:
        try:
            with open(filename, "rb") as f:
                tree = c_parser.parse(f.read())

            ast = ASTCreator.make_ast(tree.root_node)
            if args.draw_ast:
                pos = nx.drawing.nx_agraph.graphviz_layout(ast, prog="dot")
                nx.draw(
                    ast,
                    pos=pos,
                    labels={n: attr["label"] for n, attr in ast.nodes(data=True)},
                    with_labels=True,
                )
                plt.show()
            if args.write_ast is not None:
                write_dot(ast, args.write_ast)

            cfg = CFGCreator.make_cfg(ast)
            if args.draw_cfg:
                if args.each_function:
                    funcs = [
                        n
                        for n, attr in cfg.nodes(data=True)
                        if attr["label"] == "FUNC_ENTRY"
                    ]
                    for func_entry in funcs:
                        func_cfg = nx.subgraph(
                            cfg, nx.descendants(cfg, func_entry) | {func_entry}
                        )
                        draw_cfg(func_cfg, entry=func_entry)
                else:
                    draw_cfg(cfg)
            if args.write_cfg is not None:
                write_dot(cfg, args.write_cfg)
            if args.write_cfg_json:
                for n in cfg.nodes():
                    del cfg.nodes[n]["n"]
                with open(str(filename) + ".graph.json", "w") as of:
                    json.dump(json_graph.node_link_data(cfg, attrs=None), of, indent=2)
            print("successfully parsed", filename)

            duc = make_duc(cfg)
            if args.draw_duc:
                pos = nx.drawing.nx_agraph.graphviz_layout(duc, prog="dot")
                nx.draw(
                    duc,
                    pos=pos,
                    labels={n: attr["label"] for n, attr in duc.nodes(data=True)},
                    with_labels=True,
                )
                nx.draw_networkx_edge_labels(
                    duc,
                    pos=pos,
                    edge_labels={
                        (u, v): attr.get("label", "")
                        for u, v, attr in duc.edges(data=True)
                    },
                )
                plt.show()

            cpg = make_cpg(ast, cfg, duc)
            if args.draw_cpg:
                pos = nx.nx_pydot.graphviz_layout(cpg, prog="dot")
                nx.draw(cpg, pos=pos)
                nx.draw_networkx_labels(
                    cpg,
                    pos=pos,
                    labels={
                        n: attr.get("label", "<NO LABEL>")
                        for n, attr in cpg.nodes(data=True)
                    },
                )
                for graph_type, color in {
                    "AST": "black",
                    "CFG": "blue",
                    "DUC": "red",
                }.items():
                    nx.draw_networkx_edges(
                        cpg,
                        pos=pos,
                        edge_color=color,
                        edgelist=[
                            (u, v)
                            for u, v, k, attr in cpg.edges(keys=True, data=True)
                            if attr["graph_type"] == graph_type
                        ],
                    )
                nx.draw_networkx_edge_labels(
                    cpg,
                    pos=pos,
                    edge_labels={
                        (u, v): attr.get("label", "")
                        for u, v, k, attr in cpg.edges(keys=True, data=True)
                    },
                )
                plt.show()

            detect_bugs(cpg)
        except Exception:
            print("could not parse", filename)
            traceback.print_exc()
            raise



class DataflowSolver:
    """
    generic dataflow problem solver with worklist algorithm
    https://en.wikipedia.org/wiki/Data-flow_analysis

    An instance of a dataflow problem includes:
    - a CFG,
    - a domain D of "dataflow facts",
    - a dataflow fact "init" (the information true at the start of the program for forward problems, or at the end of the program for backward problems),
    - an operator ⌈⌉ (used to combine incoming information from multiple predecessors),
    - for each CFG node n, a dataflow function fn : D → D (that defines the effect of executing n). 
    """

    def __init__(self, cfg, verbose):
        self.cfg = cfg
        self.verbose = verbose

    def solve(self):
        _in = {}
        out = {}
        for n in self.cfg.nodes():
            out[n] = set()  # can optimize by OUT[n] = GEN[n];

        q = list(self.cfg.nodes())
        i = 0
        while q:
            n = q.pop(0)

            # TODO: generalize to backward problems
            _in[n] = set()
            for pred in self.cfg.predecessors(n):
                _in[n] |= out[pred]

            # TODO: generalize to backward problems
            new_out_n = self.gen(n).union(_in[n].difference(self.kill(n)))

            if self.verbose >= 2:
                print(f"{i=}, {n=}, {_in=}, {out=}, {new_out_n=}")

            if out[n] != new_out_n:
                if self.verbose >= 1:
                    print(f"{i=}, {n=} changed {out[n]} -> {new_out_n}")
                out[n] = new_out_n
                for succ in self.cfg.successors(n):
                    q.append(succ)
            i += 1

        return _in, out



from ast import Index
from matplotlib import pyplot as plt
import networkx as nx

from tree_climber.base_visitor import BaseVisitor
from tree_climber.tree_sitter_utils import c_parser


def assert_boolean_expression(n):
    assert (
        n.type.endswith("_statement")
        or n.type.endswith("_expression")
        or n.type in ("true", "false", "identifier", "number_literal")  # TODO: handle ERROR (most often shows up as comma_expression in for loop onditioanl)
    ), (n, n.type, n.text.decode())


def assert_branch_target(n):
    assert (
        n.type.endswith("_statement")
        or n.type.endswith("_expression")
        or n.type in ("else",)
    ), (n, n.type)
    

def check_ast_error_in_children(n):
    if any(c.type == "ERROR" for c in n.children):
        raise AstErrorException()


class AstErrorException(Exception):
    pass


class ASTCreator(BaseVisitor):
    """
    AST visitor which creates a CFG.
    After traversing the AST by calling visit() on the root, self.cfg has a complete CFG.
    """

    def __init__(self):
        super(ASTCreator).__init__()
        self.ast = nx.DiGraph()
        self.node_id = 0

    @staticmethod
    def make_ast(root_node):
        visitor = ASTCreator()
        visitor.visit(root_node, parent_id=None)
        return visitor.ast

    def visit(self, n, **kwargs):
        if n.type == "ERROR":
            raise AstErrorException()
        super().visit(n, **kwargs)

    def visit_for_statement(self, n, **kwargs):
        children = n.children
        check_ast_error_in_children(n)
        children = [c for c in children if c.type != "comment"]
        i = 0
        while children[i].type != "(":
            i += 1
        i += 1
        if children[i].type == "declaration":
            has_init = True
            i += 1
        elif children[i].type == ";":
            has_init = False
            i += 1
        else:
            assert children[i].type.endswith("_expression") or children[i].type in ("number_literal", "identifier"), (children[i], children[i].type)
            has_init = True
            i += 1
            assert children[i].type == ";"
            i += 1
        # pointing after semicolon
        if children[i].type == ";":
            has_cond = False
        else:
            assert_boolean_expression(children[i])
            has_cond = True
            i += 1
        # pointing at semicolon
        assert children[i].type == ";", (children[i], children[i].type, children[i].text.decode())
        i += 1
        if children[i].type == ")":
            has_incr = False
        else:
            assert children[i].type.endswith("_expression") or children[i].type in ("number_literal", "identifier"), (children[i], children[i].type)
            has_incr = True
        self.visit_default(
            n, has_init=has_init, has_cond=has_cond, has_incr=has_incr, **kwargs
        )

    def visit_case_statement(self, n, **kwargs):
        children = n.children
        check_ast_error_in_children(n)
        label_end = 0
        while children[label_end].type != ":":
            label_end += 1
        label_end -= 1
        body_begin = label_end + 2
        is_default = any(c for c in n.children if c.text.decode() == "default")
        self.visit_default(n, body_begin=body_begin, is_default=is_default, **kwargs)

    def visit_default(self, n, parent_id, **kwargs):
        code = n.text.decode()
        my_id = self.node_id
        self.node_id += 1
        if parent_id is None:
            self.ast.graph["root_node"] = my_id
        if n.is_named and n.type != "comment":

            def attr_to_label(node_type, code):
                lines = code.splitlines()
                if len(lines) > 0:
                    code = lines[0]
                    max_len = 27
                    trimmed_code = code[:max_len]
                    if len(lines) > 1 or len(code) > max_len:
                        trimmed_code += "..."
                else:
                    trimmed_code = code
                return node_type + "\n" + trimmed_code

            self.ast.add_node(
                my_id,
                n=n,
                label=attr_to_label(n.type, code),
                code=code,
                node_type=n.type,
                start=n.start_point,
                end=n.end_point,
                **kwargs
            )
            if parent_id is not None:
                self.ast.add_edge(parent_id, my_id)
            self.visit_children(n, parent_id=my_id)


def test():
    code = """#include <stdio.h>
int main()
{
    int i = 0;
    for (; i < 10; i ++) {
        printf("%d\n", i);
    }
    return 0;
}
"""
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    pos = nx.drawing.nx_agraph.graphviz_layout(ast, prog="dot")
    nx.draw(
        ast,
        pos=pos,
        labels={n: attr["label"] for n, attr in ast.nodes(data=True)},
        with_labels=True,
    )
    plt.show()



from tests.utils import *
import networkx as nx

def test_return_exclude():
    cfg = parse_and_create_cfg("""int main()
    {
        return;
        x += 5;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (3, 2)
    assert nx.is_directed_acyclic_graph(cfg)
    assert not any("x" in attr["label"] for _, attr in cfg.nodes(data=True))

def test_continue_exclude():
    cfg = parse_and_create_cfg("""int main()
    {
        for (;;) {
            continue;
            x += 5;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 1
    assert not any("x" in attr["label"] for _, attr in cfg.nodes(data=True))

def test_break_exclude():
    cfg = parse_and_create_cfg("""int main()
    {
        for (;;) {
            break;
            x += 5;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 0
    assert not any("x" in attr["label"] for _, attr in cfg.nodes(data=True))



from tests.utils import *
import networkx as nx

def test_screwy_program():
    cfg = parse_and_create_cfg("""int main()
    {
        int x = 0;
        for (int i = 0; i < 10; i ++) {
            for (int j = 0; j < 10; j ++) {
                if (true) {
                    if (false) {
                        x += j;
                    }
                    else {
                        while (true) {
                            while (false) {
                                x = 0;
                            }
                            if (x < 10) {
                                x += 5;
                            }
                            else {
                                x -= 5;
                            }
                        }
                    }
                }
            }
            x -= i;
        }
        return x;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (20, 26)
    assert len(list(nx.simple_cycles(cfg))) == 7




from tree_climber.ast_creator import ASTCreator
from tree_climber.dataflow.def_use import make_duc
from tree_climber.tree_sitter_utils import c_parser
from tree_climber.cfg_creator import CFGCreator
from tests.utils import *
import pytest


def test_get_def_use_chain():
    code = """int a = 3;
    
    int main()
    {
        int i = 0;
        int x = 0;
        end:
        x -= a;
        for (; true; ) {
            x += 5;
            if (x < 0) {
                goto end;
            }
        }
        printf("%d %d\\n", x, i);
        x = 10;
        return x;
    }
    """
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    cfg = CFGCreator.make_cfg(ast)
    duc = make_duc(cfg)

    init_x_node = get_node_by_code(duc, "int x = 0;")
    init_i_node = get_node_by_code(duc, "int i = 0;")
    true_node = get_node_by_code(duc, "true")
    a_assign_3_node = get_node_by_code(duc, "int a = 3;")
    x_minus_assign_a_node = get_node_by_code(duc, "x -= a;")
    x_plus_assign_5_node = get_node_by_code(duc, "x += 5;")
    x_assign_10_node = get_node_by_code(duc, "x = 10;")
    printf_node = get_node_by_code(cfg, """printf("%d %d\\n", x, i);""")
    return_node = get_node_by_code(cfg, "return x;")
    assert len(list(duc.predecessors(init_x_node))) == 0  # first assignment to x
    assert len(list(duc.predecessors(true_node))) == 0
    assert set(duc.predecessors(printf_node)) == {
        x_plus_assign_5_node,
        x_minus_assign_a_node,
        init_i_node,
    }  # printf can print x defined by "x = 0" or "x += 5"
    assert set(duc.predecessors(return_node)) == {
        x_assign_10_node
    }  # return is dominated by "x = 10"
    assert set(duc.predecessors(x_plus_assign_5_node)) == {
        x_plus_assign_5_node,
        x_minus_assign_a_node,
    }  # += should take into account its predecessor x
    assert set(duc.predecessors(x_minus_assign_a_node)) == {
        x_plus_assign_5_node,
        init_x_node,
        a_assign_3_node,
    }  # x -= 3 uses x += 5 (later statement) because of goto

    assert (
        duc.edges[(x_plus_assign_5_node, x_plus_assign_5_node)].get(
            "label", "<NO LABEL>"
        )
        == "x"
    )
    assert (
        duc.edges[(x_minus_assign_a_node, x_plus_assign_5_node)].get(
            "label", "<NO LABEL>"
        )
        == "x"
    )
    assert (
        duc.edges[(x_minus_assign_a_node, printf_node)].get("label", "<NO LABEL>")
        == "x"
    )
    assert duc.edges[(init_i_node, printf_node)].get("label", "<NO LABEL>") == "i"
    assert (
        duc.edges[(x_plus_assign_5_node, printf_node)].get("label", "<NO LABEL>") == "x"
    )


@pytest.mark.slow
def test_debug():
    from tests.utils import draw

    code = """int main()
    {
        int i = 0;
        int x = 0;
        end:
        x -= 3;
        for (; true; ) {
            x += 5;
            if (x < 0) {
                goto end;
            }
        }
        printf("%d %d\\n", x, i);
        x = 10;
        return x;
    }
    """
    tree = c_parser.parse(bytes(code, "utf8"))
    ast = ASTCreator.make_ast(tree.root_node)
    cfg = CFGCreator.make_cfg(ast)
    duc = make_duc(cfg)
    print(duc.nodes(data=True))

    _, ax = plt.subplots(2)
    pos = nx.drawing.nx_agraph.graphviz_layout(duc, prog="dot")
    nx.draw(
        duc,
        pos=pos,
        labels={n: attr["label"] for n, attr in duc.nodes(data=True)},
        with_labels=True,
        ax=ax[0],
    )
    nx.draw_networkx_edge_labels(
        duc,
        pos=pos,
        edge_labels={
            (u, v): attr.get("label", "") for u, v, attr in duc.edges(data=True)
        },
        ax=ax[0],
    )
    draw(cfg, ax=ax[1])



from tests.utils import *
import networkx as nx

def test_if_simple():
    cfg = parse_and_create_cfg("""int main()
    {
        if (true) {
            x += 5;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert nx.is_directed_acyclic_graph(cfg)

def test_if_nocompound():
    cfg = parse_and_create_cfg("""int main()
    {
        if (true)
            x += 5;
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert nx.is_directed_acyclic_graph(cfg)

def test_if_empty():
    cfg = parse_and_create_cfg("""int main()
    {
        if (true) {
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (3, 3)
    assert nx.is_directed_acyclic_graph(cfg)

    true_node = next(n for n, attr in cfg.nodes(data=True) if "true" in attr["label"])
    func_exit_node = next(n for n, attr in cfg.nodes(data=True) if "FUNC_EXIT" in attr["label"])
    edges_between = cfg.adj[true_node][func_exit_node]
    assert set(e.get("label", "<NO LABEL>") for e in edges_between.values()) == set(("True", "False"))

def test_if_noelse():
    cfg = parse_and_create_cfg("""int main()
    {
        if (x > 1) {
            x += 5;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert nx.is_directed_acyclic_graph(cfg)

def test_if_else():
    cfg = parse_and_create_cfg("""int main()
    {
        if (x > 1) {
            x += 5;
        }
        else {
            x += 50;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 5)
    assert nx.is_directed_acyclic_graph(cfg)

def test_if_nested():
    cfg = parse_and_create_cfg("""int main()
    {
        if (true) {
            if (false) {
                x += 5;
            }
            else {

            }
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 6)
    assert nx.is_directed_acyclic_graph(cfg)



import networkx as nx
from tree_climber.dataflow.reaching_def import ReachingDefinitionSolver


def get_uses(cfg, solver, n):
    """return the set of variables used in n"""
    used_ids = set()
    attr = cfg.nodes[n]
    if "n" in attr:
        q = [attr["n"]]
        while q:
            n = q.pop(0)
            if n.type == "identifier":
                _id = n.text.decode()
                if _id in solver.id2def.keys():
                    used_ids.add(_id)
            q.extend(n.children)
    return used_ids


def make_duc(cfg, verbose=0):
    """
       return def-use chain (DUC)

           https://textart.io/art/tag/duck/1
                                          ___
                                  ,-""   `.
                                ,'  _   e )`-._
                               /  ,' `-._<.===-'
                              /  /
                             /  ;
                 _.--.__    /   ;
    (`._    _.-""       "--'    |
    <_  `-""                     |
     <`-                          :
      (__   <__.                  ;
        `-.   '-.__.      _.'    /
           |      `-.__,-'    _,'
            `._    ,    /__,-'
               ""._|__,'< <____
                    | |  `----.`.
                    | |        | `.
                    ; |___      |-``
                    |   --<
                     `.`.<
                       `-'
    """
    duc = nx.DiGraph()
    duc.add_nodes_from(
        [(n, dict(cfg_node=n, **attr)) for n, attr in cfg.nodes(data=True)]
    )
    solver = ReachingDefinitionSolver(cfg, verbose=verbose)
    solution_in, solution_out = solver.solve()
    solution = solution_in
    for n in cfg.nodes():
        incoming_defs = solution[n]
        if len(incoming_defs) > 0:
            use_node = n

            def_ids = set(map(solver.def2id.__getitem__, incoming_defs))
            used_ids = get_uses(cfg, solver, use_node)
            used_def_ids = def_ids & used_ids
            if len(used_def_ids) > 0:
                used_defs = set.union(*map(solver.id2def.__getitem__, used_def_ids))
                used_incoming_defs = used_defs & incoming_defs

                def_nodes = set(map(solver.def2node.__getitem__, used_incoming_defs))
                for def_node in def_nodes:
                    duc.add_edge(
                        def_node,
                        use_node,
                        label=str(solver.def2id[solver.node2def[def_node]]),
                    )
    duc.remove_nodes_from(
        [
            n
            for n, attr in duc.nodes(data=True)
            if attr["label"] in ("FUNC_ENTRY", "FUNC_EXIT")
        ]
    )
    return duc



from tests.utils import *
import networkx as nx

def test_while_simple():
    cfg = parse_and_create_cfg("""int main()
    {
        while (true) {
            x = 0;
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 1

    true_node = get_node_by_code(cfg, "true")
    x_0_node = get_node_by_code(cfg, "x = 0;")
    assert get_adj_label(cfg, true_node, x_0_node) == "True"
    FUNC_EXIT_node = get_node_by_label(cfg, "FUNC_EXIT")
    assert get_adj_label(cfg, true_node, FUNC_EXIT_node) == "False"

def test_while_nested():
    cfg = parse_and_create_cfg("""int main()
    {
        while (true) {
            while (false) {
                x = 0;
            }
        }
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (5, 6)
    assert len(list(nx.simple_cycles(cfg))) == 2

def test_do_while_simple():
    cfg = parse_and_create_cfg("""int main()
    {
        do {
            x = 0;
        }
        while (true);
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (4, 4)
    assert len(list(nx.simple_cycles(cfg))) == 1
    
    true_node = get_node_by_code(cfg, "true")
    x_0_node = get_node_by_code(cfg, "x = 0;")
    assert get_adj_label(cfg, true_node, x_0_node) == "True"
    FUNC_EXIT_node = get_node_by_label(cfg, "FUNC_EXIT")
    assert get_adj_label(cfg, true_node, FUNC_EXIT_node) == "False"

def test_do_while_nested():
    cfg = parse_and_create_cfg("""int main()
    {
        do {
            while (x < 1) {
                do {
                    x = 0;
                }
                while (x < 2);
            }
        }
        while (x < 3);
    }
    """)
    assert (cfg.number_of_nodes(), cfg.number_of_edges()) == (6, 8)
    assert len(list(nx.simple_cycles(cfg))) == 3



from turtle import clone
from tree_sitter import Language, Parser
import os

from tree_climber.config import TREE_SITTER_LIB_PREFIX
from git import Repo

languages = ["c"]
language_dirs = []

for lang in languages:
    clone_dir = os.path.join(TREE_SITTER_LIB_PREFIX, "tree-sitter-c")
    if not os.path.exists(clone_dir):
        repo_url = f"https://github.com/tree-sitter/tree-sitter-{lang}.git"
        Repo.clone_from(repo_url, clone_dir)
    language_dirs.append(clone_dir)

lib_file = os.path.join(TREE_SITTER_LIB_PREFIX, "build/languages.so")
Language.build_library(
    # Store the library in the `build` directory
    lib_file,
    # Include one or more languages
    [os.path.join(TREE_SITTER_LIB_PREFIX, "tree-sitter-c"),],
)

C_LANGUAGE = Language(lib_file, "c")
c_parser = Parser()
c_parser.set_language(C_LANGUAGE)
